
"""
This module is used to define the class containing the entire Bayesian Network,
and the corresponding attributes/methods to train the model, set algorithmic options, calculate lower bound, etc.
"""

from __future__ import division
from time import time
import os
import scipy as s
import pandas as pd
import sys
import numpy.ma as ma
import math
# import resource
import numpy as np
from .nodes.variational_nodes import Variational_Node
from .nodes.multiview_nodes import Multiview_Variational_Node
from . import gpu_utils
from .utils import corr, nans, infer_platform

import warnings
warnings.filterwarnings("ignore")

class BayesNet(object):
    def __init__(self, dim, nodes):
        """ Initialisation of a Bayesian network

        PARAMETERS
        ----------
        dim: dict
            keyworded dimensionalities, ex. {'N'=10, 'M'=3, ...}
        nodes: dict
            dictionary with all nodes where the keys are the name of the node and the values are instances the 'Node' class
        """

        self.dim = dim
        self.nodes = nodes
        self.options = None

        # Training and simulations flag
        self.trained = False
        self.simulated = False

        # Set GPU mode
        # gpu_utils.gpu_mode = options['gpu_mode']

    def setTrainOptions(self, train_opts):
        """ Method to store training options """

        # Sanity checks
        assert "maxiter" in train_opts, "'maxiter' not found in the training options dictionary"
        assert "start_drop" in train_opts, "'start_drop' not found in the training options dictionary"
        assert "freq_drop" in train_opts, "'freq_drop' not found in the training options dictionary"
        assert "verbose" in train_opts, "'verbose' not found in the training options dictionary"
        assert "quiet" in train_opts, "'quiet' not found in the training options dictionary"
        # assert "tolerance" in train_opts, "'tolerance' not found in the training options dictionary"
        assert "convergence_mode" in train_opts, "'convergence_mode' not found in the training options dictionary"
        assert "forceiter" in train_opts, "'forceiter' not found in the training options dictionary"
        assert "schedule" in train_opts, "'schedule' not found in the training options dictionary"
        assert "start_sparsity" in train_opts, "'start_sparsity' not found in the training options dictionary"
        assert "gpu_mode" in train_opts, "'gpu_mode' not found in the training options dictionary"
        assert "start_elbo" in train_opts, "'gpu_mode' not found in the training options dictionary"

        self.options = train_opts

    def getParameters(self, *nodes):
        """ Method to collect all parameters of a given set of nodes

        PARAMETERS
        ----------
        nodes: iterable
            name of the nodes (all nodes by default)
        """

        if len(nodes) == 0: nodes = self.nodes.keys()
        params = {}
        for node in nodes:
            tmp = self.nodes[node].getParameters()
            if tmp != None: params[node] = tmp
        return params

    def getExpectations(self, only_first_moments=False, *nodes):
        """Method to collect all expectations of a given set of nodes

        PARAMETERS
        ----------
        only_first_moments: bool
            get only first moments? (Default is False)
        nodes: list
            name of the nodes (Default inp.all nodes)
        """

        if len(nodes) == 0: nodes = self.nodes.keys()
        expectations = {}
        for node in nodes:
            if only_first_moments:
                tmp = self.nodes[node].getExpectation()
            else:
                tmp = self.nodes[node].getExpectations()
            expectations[node] = tmp
        return expectations

    def getNodes(self):
        """ Method to return all nodes """
        return self.nodes

    def calculate_variance_explained(self, total=False):

        # Collect relevant expectations
        Z = self.nodes['Z'].getExpectation()
        W = self.nodes["W"].getExpectation()
        Y = self.nodes["Y"].getExpectation()

        # Get groups
        groups = self.nodes["Y"].nodes[0].groups

        # to maintain correct ordering of groups in R2
        unique_groups, idx = np.unique(groups, return_index=True)
        unique_groups = unique_groups[np.argsort(idx)]

        if total:
            r2 = [ np.zeros(self.dim['M']) for g in range(self.dim['G'])]
        else:
            r2 = [ np.zeros([self.dim['M'], self.dim['K']])  for g in range(self.dim['G'])]

        for m in range(self.dim['M']):
            mask = self.nodes["Y"].getNodes()[m].getMask(full=True)
            for g in range(self.dim['G']):
                gidx = unique_groups[g]
                gg = groups==gidx
                SS = np.square(Y[m][gg,:]).sum()

                # Total variance explained (using all factors)
                if total:
                    Ypred = np.dot(Z[gg,:], W[m].T)
                    Ypred[mask[gg,:]] = 0.
                    Res = np.sum((Y[m][gg, :] - Ypred) ** 2.)
                    r2[g][m] = 1. - Res / SS

                # Variance explained per factor
                else:
                    for k in range(self.dim['K']):
                        Ypred = s.outer(Z[gg,k], W[m][:,k])
                        Ypred[mask[gg,:]] = 0.
                        Res = np.sum((Y[m][gg,:] - Ypred)**2.)
                        r2[g][m,k] = 1. - Res/SS
        return r2

    def removeInactiveFactors(self, min_r2=None, return_idx = False):
        """Method to remove inactive factors

        PARAMETERS
        ----------
        min_r2: float
            threshold to shut down factors based on a minimum variance explained per group and view
        """
        drop_dic = {}

        if min_r2 is not None:
            r2 = self.calculate_variance_explained()

            tmp = [ s.where( (r2[g]>min_r2).sum(axis=0) == 0)[0] for g in range(self.dim['G']) ]
            drop_dic["min_r2"] = list(set.intersection(*map(set,tmp)))
            if len(drop_dic["min_r2"]) > 0:
                drop_dic["min_r2"] = [ np.random.choice(drop_dic["min_r2"]) ]

        # Drop the factors
        drop = s.unique(np.concatenate(list(drop_dic.values())))
        if len(drop) > 0:
            for node in self.nodes.keys():
                self.nodes[node].removeFactors(drop)
        self.dim['K'] -= len(drop)

        # Remove factor-wise training stats
        # if len(drop) > 0:
            # if 'Sigma' in self.nodes.keys():
            #     self.lscales = self.lscales.drop(columns = drop)
            #     self.lscales.columns = range(0, len(self.lscales.columns))
            #     self.scales = self.scales.drop(columns = drop)
            #     self.scales.columns = range(0, len(self.scales.columns))
        if self.dim['K']==0:
            print("All factors shut down, no structure found in the data.")
            exit()

        if return_idx: return drop

        pass

    def precompute(self):
        # Precompute terms
        for n in self.nodes:
            self.nodes[n].precompute(self.options)

        # Precompute ELBO
        for node in self.nodes["Y"].getNodes(): node.TauTrick = False # important to do this for ELBO computation
        elbo = self.calculateELBO()
        for node in self.nodes["Y"].getNodes(): node.TauTrick = True

        if self.options['verbose']:
            print("ELBO before training:")
            print("".join([ "%s=%.2f  " % (k,v) for k,v in elbo.drop("total").iteritems() ]) + "\nTotal: %.2f\n" % elbo["total"])
        else:
            if not self.options['quiet']:
                print('ELBO before training: %.2f \n' % elbo["total"])

        return elbo

    def iterate(self):
        """Method to start iterating and updating the variables using the VB algorithm"""

        # Define some variables to monitor training
        nodes = list(self.getVariationalNodes().keys())
        elbo = pd.DataFrame(data = nans((self.options['maxiter']+1, len(nodes)+1 )), columns = nodes+["total"] )
        number_factors = nans((self.options['maxiter']+1))
        iter_time = nans((self.options['maxiter']+1))
        # keep track of factor-wise training statistics (attribute as needs to be accounted for in factor dropping)
        # if 'Sigma' in self.nodes.keys():
        #     self.lscales = pd.DataFrame(data = nans((self.options['maxiter'], self.dim['K'])), columns = range(self.dim['K']))
        #     self.scales = pd.DataFrame(data = nans((self.options['maxiter'], self.dim['K'])), columns = range(self.dim['K']))

        # Precompute
        converged = False; convergence_token = 1
        elbo.iloc[0] = self.precompute()
        number_factors[0] = self.dim['K']
        iter_time[0] = 0.

        try:
            for i in range(1,self.options['maxiter']):
                t = time();

                # Remove inactive factors
                if (i>=self.options["start_drop"]) and (i%self.options['freq_drop']) == 0:
                    if self.options['drop']["min_r2"] is not None:
                        self.removeInactiveFactors(**self.options['drop'])
                    number_factors[i] = self.dim["K"]

                # Update node by node, with E and M step merged
                t_updates = time()
                for node in self.options['schedule']:
                    if (node=="ThetaW" or node=="ThetaZ") and i<self.options['start_sparsity']:
                        continue
                    self.nodes[node].update()
                t_updates = time() - t_updates

                # Calculate Evidence Lower Bound
                if (i>=self.options["start_elbo"]) and ((i-self.options["start_elbo"])%self.options['freqELBO']==0):
                    t_elbo = time()
                    elbo.iloc[i] = self.calculateELBO()
                    t_elbo = time() - t_elbo

                    # Check convergence using the ELBO
                    if i==self.options["start_elbo"]: 
                        delta_elbo = elbo.iloc[i]["total"]-elbo.iloc[0]["total"]
                    else:
                        delta_elbo = elbo.iloc[i]["total"]-elbo.iloc[i-self.options['freqELBO']]["total"]

                    # Print ELBO monitoring
                    if not self.options['quiet']:
                        print("Iteration %d: time=%.2f, ELBO=%.2f, deltaELBO=%.3f (%.8f%%), Factors=%d" % (i, time()-t, elbo.iloc[i]["total"], delta_elbo, 100*abs(delta_elbo/elbo.iloc[0]["total"]), (self.dim['K'])))
                        if delta_elbo<0 and not self.options['stochastic']: print("Warning, lower bound is decreasing...\a")

                    # Print ELBO decomposed by node and variance explained
                    if self.options['verbose']:
                        print("- ELBO decomposition:  " + "".join([ "%s=%.2f  " % (k,v) for k,v in elbo.iloc[i].drop("total").iteritems() ]))
                        print('- Time spent in ELBO computation: %.1f%%' % (100*t_elbo/(t_updates+t_elbo)) )

                    # Assess convergence
                    if i>self.options["start_elbo"] and i>self.options["min_iter"] and not self.options['forceiter']:
                        convergence_token, converged = self.assess_convergence(delta_elbo, elbo.iloc[0]["total"], convergence_token)
                        if converged:
                            number_factors = number_factors[:i]
                            elbo = elbo[:i]
                            iter_time = iter_time[:i]
                            print ("\nConverged!\n"); break

                # Do not calculate lower bound
                else:
                    if not self.options['quiet']: print("Iteration %d: time=%.2f, Factors=%d" % (i,time()-t,self.dim["K"]))

                # Print other statistics
                if self.options['verbose']:
                    self.print_verbose_message(i)


                iter_time[i] = time()-t
                
                # Flush (we need this to print when running on the cluster)
                sys.stdout.flush()

            
            self.trained = True

        except KeyboardInterrupt:
            self.trained = False

        finally:
            # Finish by collecting the training statistics
            self.train_stats = { 'time':iter_time, 'number_factors':number_factors, 'elbo':elbo["total"].values, 'elbo_terms':elbo.drop("total",1) }
            if 'Sigma' in self.nodes.keys():
                tmp = self.nodes['Sigma'].getParameters() # save only last iteration
                self.train_stats['length_scales'] = tmp['l']
                self.train_stats['scales'] = tmp['scale']
                self.train_stats['Kg'] = tmp['Kg']

                # self.train_stats['length_scales'] = self.lscales
                # self.train_stats['scales'] = self.scales

    def print_verbose_message(self, i):
        """Method to print training statistics if Verbose is TRUE"""

        # Memory usage (does not work in Windows)
        # print('Peak memory usage: %.2f MB' % (resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / infer_platform() ))

        # Variance explained
        r2 = s.asarray(self.calculate_variance_explained(total=True)).mean(axis=0)
        r2[r2<0] = 0.
        print("- Variance explained:  " + "   ".join([ "View %s: %.2f%%" % (m,100*r2[m]) for m in range(self.dim["M"])]))

        # Sparsity levels of the weights
        W = self.nodes["W"].getExpectation()
        foo = [np.mean(np.absolute(W[m])<1e-3) for m in range(self.dim["M"])]
        print("- Fraction of zero weights:  " + "   ".join([ "View %s: %.0f%%" % (m,100*foo[m]) for m in range(self.dim["M"])]))

        # Correlation between factors
        Z = self.nodes["Z"].getExpectation()
        Z += np.random.normal(np.zeros(Z.shape),1e-10)
        r = np.absolute(corr(Z.T,Z.T)); np.fill_diagonal(r,0)
        print("- Maximum correlation between factors: %.2f" % (np.nanmax(r)))

        # Factor norm
        bar = np.mean(np.square(Z),axis=0)
        print("- Factor norms:  " + " ".join([ "%.2f" % bar[k] for k in range(Z.shape[1])]))

        # Tau
        tau = self.nodes["Tau"].getExpectation()
        print("- Tau per view (average):  " + "   ".join([ "View %s: %.2f" % (m,tau[m].mean()) for m in range(self.dim["M"])]))

        #Sigma:
        if 'Sigma' in self.nodes.keys():
            sigma = self.nodes["Sigma"]
            if i >= sigma.start_opt and i % sigma.opt_freq == 0:
                print('Sigma node has been optimised:\n- Lengthscales = %s \n- Scale = %s' % \
                (np.array2string(sigma.get_ls(), precision=2, separator=", "), np.array2string(1-sigma.get_zeta(), precision=2, separator=", ")))

        print("\n")

    def assess_convergence(self, delta_elbo, first_elbo, convergence_token):
        converged = False

        # Option 1: deltaELBO
        # if abs(delta_elbo) < self.options['tolerance']: 
        #     converged = True

        # Assess convergence based on the fraction of deltaELBO change
        if self.options["convergence_mode"] == "fast":
            convergence_threshold = 0.0005
        elif self.options["convergence_mode"] == "medium":
            convergence_threshold = 0.00005
        elif self.options["convergence_mode"] == "slow":
            convergence_threshold = 0.000005
        else:
            print("Convergence mode not recognised"); exit()

        if 100*abs(delta_elbo/first_elbo) < convergence_threshold: 
            convergence_token += 1
            if convergence_token==3: converged = True
        else:
            convergence_token = 1

        return convergence_token, converged

    def getVariationalNodes(self):
        """ Method to return all variational nodes """
        # TODO problem with dictionnary comprehension here
        to_ret = {}
        for node in self.nodes.keys():
            if isinstance(self.nodes[node],Variational_Node):
                to_ret[node] =self.nodes[node]

        return to_ret
        # return { node:self.nodes[node] for node in self.nodes.keys() if isinstance(self.nodes[node],Variational_Node)}
        # return { k:v for k,v in self.nodes.items() if isinstance(v,Variational_Node) }

    def getTrainingStats(self):
        """ Method to return training statistics """
        return self.train_stats

    def getTrainingOpts(self):
        """ Method to return training options """
        return self.options

    def getTrainingData(self):
        """ Method to return training data """
        return self.nodes["Y"].getValues()

    def calculateELBO(self, *nodes):
        """Method to calculate the Evidence Lower Bound of the model"""

        weights = [1] * self.dim['M']        
        if self.options['weight_views'] and self.dim['M'] > 1:
            total_w = np.asarray(self.dim['D']).sum()
            weights = np.asarray([total_w / (self.dim['M'] * self.dim['D'][m]) for m in range(self.dim['M'])])
            weights = weights / weights.sum() * self.dim['M']
            # weights = [(total_w-self.dim['D'][m])/total_w * self.dim['M'] / (self.dim['M'] - 1)  for m in range(self.dim['M'])]
            

        if len(nodes) == 0: nodes = self.getVariationalNodes().keys()
        elbo = pd.Series(np.zeros(len(nodes)+1), index=list(nodes)+["total"])
        for node in nodes:
            if isinstance(self.nodes[node], Multiview_Variational_Node):
                elbo[node] = float(self.nodes[node].calculateELBO(weights = weights))
            else:
                elbo[node] = float(self.nodes[node].calculateELBO())
            elbo["total"] += elbo[node]
        return elbo


class StochasticBayesNet(BayesNet):
    def __init__(self, dim, nodes):
        super().__init__(dim=dim, nodes=nodes)

    def step_size(self, i):
        # return the step size for the considered iteration
        return (i + self.options['learning_rate'])**(-self.options['forgetting_rate'])

    def step_size2(self, i):
        # return the step size for the considered iteration
        return self.options['learning_rate'] / ((1 + self.options['forgetting_rate'] * i)**(3./4.))

    def sample_mini_batch(self):
        """ Method to define mini batches"""
        S = int( self.options['batch_size'] * self.dim['N'] )
        ix = np.random.choice(range(self.dim['N']), size=S, replace=False)
        self.define_mini_batch(ix)
        return ix

    def sample_mini_batch_no_replace(self, i):
        """ Method to define mini batches"""

        i -= 1 # This is because we start at iteration 1 in the main loop

        # Sample mini-batch indices and define epoch
        n_batches = math.ceil(1./self.options['batch_size'])
        S = self.options['batch_size'] * self.dim['N']
        batch_ix = i % n_batches
        epoch = int(i / n_batches)
        if batch_ix == 0:
            print("\n## Epoch %s ##" % str(epoch+1))
            print("-------------------------------------------------------------------------------------------")
            self.shuffled_ix = np.random.choice(range(self.dim['N']), size=self.dim['N'], replace=False)

        min = int(S * batch_ix)
        max = int(S * (batch_ix + 1))

        if max>self.dim['N']: print("Error in stochastic"); exit()

        # Define mini batch
        ix = self.shuffled_ix[min:max]
        self.define_mini_batch(ix)

        return ix, epoch
    
    def define_mini_batch(self, ix):
        # Define mini-batch for each node
        self.nodes['Z'].define_mini_batch(ix)
        self.nodes['Y'].define_mini_batch(ix)
        self.nodes['Tau'].define_mini_batch(ix)
        if 'AlphaZ' in self.nodes:
            self.nodes['AlphaZ'].define_mini_batch(ix)
        if 'ThetaZ' in self.nodes:
            self.nodes['ThetaZ'].define_mini_batch(ix)  
        if 'Sigma' in self.nodes:
            self.nodes['Sigma'].define_mini_batch(ix)
  

    def iterate(self):
        """Method to start iterating and updating the variables using the VB algorithm"""

        # Define some variables to monitor training
        nodes = list(self.getVariationalNodes().keys())
        elbo = pd.DataFrame(data = nans((self.options['maxiter']+1, len(nodes)+1 )), columns = nodes+["total"] )
        number_factors = nans((self.options['maxiter']+1))
        iter_time = nans((self.options['maxiter']+1))
        # if 'Sigma' in self.nodes.keys():
        #     self.lscales = pd.DataFrame(data = nans((self.options['maxiter'], self.dim['K'])), columns = range(self.dim['K']))
        #     self.scales = pd.DataFrame(data = nans((self.options['maxiter'], self.dim['K'])), columns = range(self.dim['K']))

        # Precompute
        converged = False; convergence_token = 1
        elbo.iloc[0] = self.precompute()
        number_factors[0] = self.dim['K']
        iter_time[0] = 0.
        iter_count = 0

        # Print stochastic settings before training
        print("Using stochastic variational inference with the following parameters:")
        print("- Batch size (fraction of samples): %.2f\n- Forgetting rate: %.2f\n- Learning rate: %.2f\n- Starts at iteration: %d \n" % 
            (100*self.options['batch_size'], self.options['forgetting_rate'], self.options['learning_rate'], self.options['start_stochastic']) )
        ix = None

        for i in range(1, self.options['maxiter']):
            t = time();

            # Sample mini-batch and define step size for stochastic inference
            if i>=self.options["start_stochastic"]:
                ix, epoch = self.sample_mini_batch_no_replace(i-(self.options["start_stochastic"]-1))
                ro = self.step_size2(epoch)
            else:
                ro = 1.

            # Doesn't really make a big difference...
            # if i==self.options["start_stochastic"]:
            #     self.options['schedule'].pop( self.options['schedule'].index("Z") )
            #     self.options['schedule'].insert(1,"Z")

            # Remove inactive factors
            if (i>=self.options["start_drop"]) and (i%self.options['freq_drop']) == 0:
                if self.options['drop']["min_r2"] is not None:
                    self.removeInactiveFactors(**self.options['drop'])
                number_factors[i] = self.dim["K"]

            # Update node by node, with E and M step merged
            t_updates = time()
            for node in self.options['schedule']:
                if (node=="ThetaW" or node=="ThetaZ") and i<self.options['start_sparsity']:
                    continue
                self.nodes[node].update(ix, ro)
            t_updates = time() - t_updates

            # # Save lengthscales from Sigma node
            # if 'Sigma' in self.nodes.keys():
            #     tmp = self.nodes['Sigma'].getParameters()
            #     self.lscales.iloc[i] = tmp['l']
            #     self.scales.iloc[i] = tmp['scale']
            #     self.Kg.iloc[i] = tmp['Kg']

            # Calculate Evidence Lower Bound
            if (i>=self.options["start_elbo"]) and ((i-self.options["start_elbo"])%self.options['freqELBO']==0):
                t_elbo = time()
                elbo.iloc[i] = self.calculateELBO()
                t_elbo = time() - t_elbo

                # Check convergence using the ELBO
                if i==self.options["start_elbo"]: 
                    delta_elbo = elbo.iloc[i]["total"]-elbo.iloc[0]["total"]
                else:
                    delta_elbo = elbo.iloc[i]["total"]-elbo.iloc[i-self.options['freqELBO']]["total"]

                # Print ELBO monitoring
                print("Iteration %d: time=%.2f, ELBO=%.2f, deltaELBO=%.3f (%.9f%%), Factors=%d" % (i, time()-t, elbo.iloc[i]["total"], delta_elbo, 100*abs(delta_elbo/elbo.iloc[0]["total"]), (self.dim['K'])))
                if delta_elbo<0 and not self.options['stochastic']: print("Warning, lower bound is decreasing...\a")

                # Print ELBO decomposed by node and variance explained
                if self.options['verbose']:
                    print("- ELBO decomposition:  " + "".join([ "%s=%.2f  " % (k,v) for k,v in elbo.iloc[i].drop("total").iteritems() ]))
                    print('- Time spent in ELBO computation: %.1f%%' % (100*t_elbo/(t_updates+t_elbo)) )

                # Assess convergence
                if i>self.options["start_elbo"] and not self.options['forceiter']:
                    convergence_token, converged = self.assess_convergence(delta_elbo, elbo.iloc[0]["total"], convergence_token)
                    if converged:
                        number_factors = number_factors[:i]
                        elbo = elbo[:i]
                        iter_time = iter_time[:i]
                        print("\nConverged!\n"); break

            # Do not calculate lower bound
            else:
                print("Iteration %d: time=%.2f, Factors=%d" % (i,time()-t,self.dim["K"]))

            # Print other statistics
            if i>=(self.options["start_stochastic"]):
                print("- Step size: %.3f" % ro )

            if self.options['verbose']:
                self.print_verbose_message(i)
            # print("")

            iter_time[i] = time()-t
            iter_count += 1
            
            # Flush (we need this to print when running on the cluster)
            sys.stdout.flush()

        if iter_count+1 == self.options['maxiter']:
            print("\nMaximum number of iterations reached: {}\n".format(self.options['maxiter']))

        # Finish by collecting the training statistics
        self.train_stats = { 'time':iter_time, 'number_factors':number_factors, 'elbo':elbo["total"].values, 'elbo_terms':elbo.drop("total",1) }
        if 'Sigma' in self.nodes.keys():
            tmp = self.nodes['Sigma'].getParameters() # save only last iteration
            self.train_stats['length_scales'] = tmp['l']
            self.train_stats['scales'] = tmp['scale']
            self.train_stats['Kg'] = tmp['Kg']

            # self.train_stats['length_scales'] = self.lscales
            # self.train_stats['scales'] = self.scales

        self.trained = True



r"""
bulk (A omic framework for bulk omic analysis)
"""

#from Pyomic.bulk.DeGene import find_DEG,Density_norm,Plot_gene_expression,ID_mapping
#from Pyomic.bulk.Gene_module import pywgcna

from ._Gene_module import pyWGCNA
from ._Enrichment import pyGSEA,pyGSE,geneset_enrichment,geneset_plot,geneset_enrichment_GSEA
from ._network import pyPPI,string_interaction,string_map,generate_G
from ._chm13 import get_chm13_gene,find_chm13_gene
from ._Deseq2 import pyDEG,deseq2_normalize,estimateSizeFactors,estimateDispersions,Matrix_ID_mapping,data_drop_duplicates_index
from ._tcga import pyTCGA
from ._combat import batch_correction


from typing import Union, List, Dict
import numpy as np
import pandas as pd
import scanpy as sc


def _validate_obsm_key(ad, key, as_df=True):
    """
    Validates and retrieves the data associated with a specified key from the provided AnnData object.

    Parameters
    ----------
    ad : sc.AnnData
        The annotated data matrix from which the data is to be retrieved.
    key : str
        The key for accessing the data from the AnnData object's obsm.
    as_df : bool, optional
        If True, the data will be returned as pandas DataFrame with pseudotime as column names.
        If False, the data will be returned as numpy array.
        Default is True.

    Returns
    -------
    data : pd.DataFrame
        A DataFrame containing the data associated with the specified key.
    data_names : List[str]
        A list of column names for the DataFrame.

    Raises
    ------
    KeyError
        If the key or its corresponding columns are not found in the AnnData object.
    """
    if key not in ad.obsm:
        raise KeyError(f"{key} not found in ad.obsm")
    data = ad.obsm[key]
    if not isinstance(data, pd.DataFrame):
        if key + "_columns" not in ad.uns:
            raise KeyError(
                f"{key}_columns not found in ad.uns and ad.obsm[key] is not a DataFrame."
            )
        data_names = list(ad.uns[key + "_columns"])
        if as_df:
            data = pd.DataFrame(data, columns=data_names, index=ad.obs_names)
    else:
        data_names = list(data.columns)
        if not as_df:
            data = data.values
    return data, data_names


def _validate_varm_key(ad, key, as_df=True):
    """
    Validates and retrieves the data associated with a specified key from the provided AnnData object's varm attribute.

    Parameters
    ----------
    ad : sc.AnnData
        The annotated data matrix from which the data is to be retrieved.
    key : str
        The key for accessing the data from the AnnData object's varm.
    as_df : bool, optional
        If True, the trends will be returned as pandas DataFrame with pseudotime as column names.
        If False, the trends will be returned as numpy array.
        Default is True.

    Returns
    -------
    data : Union[pd.DataFrame, np.ndarray]
        A DataFrame or numpy array containing the data associated with the specified key.
    data_names : np.ndarray
        A an array of pseudotimes.

    Raises
    ------
    KeyError
        If the key or its corresponding columns are not found in the AnnData object.
    """
    if key not in ad.varm:
        raise KeyError(f"{key} not found in ad.varm")
    data = ad.varm[key]
    if not isinstance(data, pd.DataFrame):
        if key + "_pseudotime" not in ad.uns:
            raise KeyError(
                f"{key}_pseudotime not found in ad.uns and ad.varm[key] is not a DataFrame."
            )
        data_names = np.array(ad.uns[key + "_pseudotime"])
        if as_df:
            data = pd.DataFrame(data, columns=data_names, index=ad.var_names)
    else:
        data_names = np.array(data.columns.astype(float))
        if not as_df:
            data = data.values
    return data, data_names


def _validate_gene_trend_input(
    data: Union[sc.AnnData, Dict],
    gene_trend_key: str = "gene_trends",
    branch_names: Union[str, List[str]] = "branch_masks",
) -> Dict:
    """
    Validates the input for gene trend plots, and converts it into a dictionary of gene trends.

    Parameters
    ----------
    data : Union[sc.AnnData, Dict]
        An AnnData object or a dictionary containing gene trends.
    gene_trend_key : str, optional
        Key to access gene trends in the varm of the AnnData object. Default is 'gene_trends'.
    branch_names : Union[str, List[str]], optional
        Key to retrieve branch names from the AnnData object or a list of branch names. If a string is provided,
        it is assumed to be a key in AnnData.uns. Default is 'branch_masks'.

    Returns
    -------
    gene_trends : Dict
        A dictionary containing gene trends.

    Raises
    ------
    KeyError
        If 'branch_names' is a string that is not found in .uns, or if 'gene_trend_key + "_" + branch_name'
        is not found in .varm.
    ValueError
        If 'data' is neither an AnnData object nor a dictionary.
    """
    if isinstance(data, sc.AnnData):
        if isinstance(branch_names, str):
            if branch_names in data.uns.keys():
                branch_names = data.uns[branch_names]
            elif branch_names in data.obsm.keys() and isinstance(
                data.obsm[branch_names], pd.DataFrame
            ):
                branch_names = list(data.obsm[branch_names].columns)
            elif branch_names + "_columns" in data.uns.keys():
                branch_names = data.uns[branch_names + "_columns"]
            else:
                raise KeyError(
                    f"The provided key '{branch_names}' is not found in AnnData.uns or as a DataFrame in AnnData.obsm. "
                    "Please ensure the 'branch_names' either exists in AnnData.uns or is a list of branch names."
                )

        gene_trends = dict()
        for branch in branch_names:
            trends, pt_grid = _validate_varm_key(data, gene_trend_key + "_" + branch)
            gene_trends[branch] = {"trends": trends}
    elif isinstance(data, Dict):
        gene_trends = data
    else:
        raise ValueError(
            "The input 'data' must be an instance of either AnnData object or dictionary."
        )

    return gene_trends



import pandas as pd
import numpy as np
import scanpy as sc

tg_install=False

class Tangram(object):
    
    def check_tangram(self):
        """
        
        """
        global tg_install
        try:
            import tangram as tg
            tg_install=True
            print('tangram have been install version:',tg.__version__)
        except ImportError:
            raise ImportError(
                'Please install the tangram: `pip install -U tangram-sc`.'
            )
    
    def __init__(self,adata_sc,adata_sp,clusters,marker_size=100,
                gene_to_lowercase=False):
        
        self.check_tangram()
        global tg_install
        if tg_install==True:
            global_imports("tangram","tg")
        
        
        ad_map_dict={}

        #adata_sc=adata_raw_all
        adata_sc.uns['log1p']={}
        adata_sc.uns['log1p']['base']=None
        sc.pp.filter_genes(adata_sc, min_cells=1)
        sc.tl.rank_genes_groups(adata_sc, groupby=clusters, key_added=f'{clusters}_rank_genes_groups',use_raw=False)
        markers_df = pd.DataFrame(adata_sc.uns[f"{clusters}_rank_genes_groups"]["names"]).iloc[0:marker_size, :]
        
        markers = list(np.unique(markers_df.melt().value.values))
        print('...Calculate The Number of Markers:',len(markers))
        
        self.adata_sc=adata_sc
        self.adata_sp=adata_sp
        self.clusters=clusters
        self.markers=markers
        
        #import tangram as tg
        tg.pp_adatas(self.adata_sc, self.adata_sp, genes=self.markers,gene_to_lowercase=gene_to_lowercase)
        
        print('...Model prepared successfully')
        
    def train(self,mode="clusters",num_epochs=500,device="cuda:0",**kwargs):

        ad_map = tg.map_cells_to_space(self.adata_sc, self.adata_sp,
            #mode="cells",
            mode=mode,
            cluster_label=self.clusters,  # .obs field w cell types
            #density_prior='rna_count_based',
            num_epochs=num_epochs,
            device=device,
            **kwargs
            #device='cpu',
        )
        
    
        tg.project_cell_annotations(ad_map, self.adata_sp, annotation=self.clusters)
        print('...Model train successfully')
        
    def cell2location(self,annotation_list=None):
        adata_plot=self.adata_sp.copy()
        # construct df_plot
        if annotation_list==None:
            annotation_list=list(set(self.adata_sc.obs[self.clusters]))
        
        df = adata_plot.obsm["tangram_ct_pred"][annotation_list]
        construct_obs_plot(df, adata_plot, perc=0)
        return adata_plot
        
        

def construct_obs_plot(df_plot, adata, perc=0, suffix=None):
    # clip
    df_plot = df_plot.clip(df_plot.quantile(perc), df_plot.quantile(1 - perc), axis=1)

    # normalize
    df_plot = (df_plot - df_plot.min()) / (df_plot.max() - df_plot.min())

    if suffix:
        df_plot = df_plot.add_suffix(" ({})".format(suffix))
    adata.obs = pd.concat([adata.obs, df_plot], axis=1)

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)


#!/usr/bin/env python
"""
# Author: Kangning Dong
# File Name: __init__.py
# Description:
"""

__author__ = "Kangning Dong"
__email__ = "dongkangning16@mails.ucas.ac.cn"

from .STAGATE import STAGATE
from .Train_STAGATE import train_STAGATE
from .utils import Transfer_pytorch_Data, Cal_Spatial_Net, Stats_Spatial_Net, mclust_R, Cal_Spatial_Net_3D, Batch_Data



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as td


__all__ = [
    'BerpoDecoder',
]


class BernoulliDecoder(nn.Module):
    def __init__(self, num_nodes, num_edges, balance_loss=False):
        """Base class for Bernoulli decoder.

        Args:
            num_nodes: Number of nodes in a graph.
            num_edges: Number of edges in a graph.
            balance_loss: Whether to balance contribution from edges and non-edges.
        """
        super().__init__()
        self.num_nodes = num_nodes
        self.num_edges = num_edges
        self.num_possible_edges = num_nodes**2 - num_nodes
        self.num_nonedges = self.num_possible_edges - self.num_edges
        self.balance_loss = balance_loss

    def forward_batch(self, emb, idx):
        """Compute probabilities of given edges.

        Args:
            emb: embedding matrix, shape (num_nodes, emb_dim)
            idx: edge indices, shape (batch_size, 2)

        Returns:
            edge_probs: Bernoulli distribution for given edges, shape (batch_size)
        """
        raise NotImplementedError

    def forward_full(self, emb):
        """Compute probabilities for all edges.

        Args:
            emb: embedding matrix, shape (num_nodes, emb_dim)

        Returns:
            edge_probs: Bernoulli distribution for all edges, shape (num_nodes, num_nodes)
        """
        raise NotImplementedError

    def loss_batch(self, emb, ones_idx, zeros_idx):
        """Compute loss for given edges and non-edges."""
        raise NotImplementedError

    def loss_full(self, emb, adj):
        """Compute loss for all edges and non-edges."""
        raise NotImplementedError


class BerpoDecoder(BernoulliDecoder):
    def __init__(self, num_nodes, num_edges, balance_loss=False):
        super().__init__(num_nodes, num_edges, balance_loss)
        edge_proba = num_edges / (num_nodes**2 - num_nodes)
        self.eps = -np.log(1 - edge_proba)

    def forward_batch(self, emb, idx):
        """Compute probabilities of given edges.

        Args:
            emb: embedding matrix, shape (num_nodes, emb_dim)
            idx: edge indices, shape (batch_size, 2)

        Returns:
            edge_probs: Bernoulli distribution for given edges, shape (batch_size)
        """
        e1, e2 = idx.t()
        logits = torch.sum(emb[e1] * emb[e2], dim=1)
        logits += self.eps
        probs = 1 - torch.exp(-logits)
        return td.Bernoulli(probs=probs)

    def forward_full(self, emb):
        """Compute probabilities for all edges.

        Args:
            emb: embedding matrix, shape (num_nodes, emb_dim)

        Returns:
            edge_probs: Bernoulli distribution for all edges, shape (num_nodes, num_nodes)
        """
        logits = emb @ emb.t()
        logits += self.eps
        probs = 1 - torch.exp(-logits)
        return td.Bernoulli(probs=probs)

    def loss_batch(self, emb, ones_idx, zeros_idx):
        """Compute BerPo loss for a batch of edges and non-edges."""
        # Loss for edges
        e1, e2 = ones_idx[:, 0], ones_idx[:, 1]
        edge_dots = torch.sum(emb[e1] * emb[e2], dim=1)
        loss_edges = -torch.mean(torch.log(-torch.expm1(-self.eps - edge_dots)))

        # Loss for non-edges
        ne1, ne2 = zeros_idx[:, 0], zeros_idx[:, 1]
        loss_nonedges = torch.mean(torch.sum(emb[ne1] * emb[ne2], dim=1))
        if self.balance_loss:
            neg_scale = 1.0
        else:
            neg_scale = self.num_nonedges / self.num_edges
        return (loss_edges + neg_scale * loss_nonedges) / (1 + neg_scale)

    def loss_full(self, emb, adj):
        """Compute BerPo loss for all edges & non-edges in a graph."""
        e1, e2 = adj.nonzero()
        edge_dots = torch.sum(emb[e1] * emb[e2], dim=1)
        loss_edges = -torch.sum(torch.log(-torch.expm1(-self.eps - edge_dots)))

        # Correct for overcounting F_u * F_v for edges and nodes with themselves
        self_dots_sum = torch.sum(emb * emb)
        correction = self_dots_sum + torch.sum(edge_dots)
        sum_emb = torch.sum(emb, dim=0, keepdim=True).t()
        loss_nonedges = torch.sum(emb @ sum_emb) - correction

        if self.balance_loss:
            neg_scale = 1.0
        else:
            neg_scale = self.num_nonedges / self.num_edges
        return (loss_edges / self.num_edges + neg_scale * loss_nonedges / self.num_nonedges) / (1 + neg_scale)
